#include <petsc.h>

/* Define context for user-provided preconditioner */
typedef struct {
  KSP kspSchurMp, kspSchurLp;
} SampleShellPC;

/* Declare routines for user-provided preconditioner */
extern PetscErrorCode SampleShellPCCreate(SampleShellPC**, Mat, Mat);
extern PetscErrorCode SampleShellPCApply(PC,Vec x,Vec y);
extern PetscErrorCode SampleShellPCDestroy(PC);

#undef __FUNCT__
#define __FUNCT__ "main"
int main(int argc,char **argv) {
  PetscInitialize(&argc,&argv,(char*)0,NULL);
  Mat            Mp,Lp,Sp;             /* matrices */
  KSP            ksp;
  char           filename[PETSC_MAX_PATH_LEN];
  PetscViewer    viewer;
  PetscBool      flg;
  PetscErrorCode ierr;

  PC         pc;
  KSPType    ksptype;
  PCType     pctype;

  SampleShellPC *shell;

  ierr = PetscOptionsGetString(NULL,NULL,"-fSp",filename,PETSC_MAX_PATH_LEN,&flg);CHKERRQ(ierr);
  if (!flg) SETERRQ(PETSC_COMM_WORLD,1,"Must indicate a file name for matrix Sp with the -fSp option");
  ierr = PetscViewerBinaryOpen(PETSC_COMM_WORLD,filename,FILE_MODE_READ,&viewer);CHKERRQ(ierr);
  ierr = MatCreate(PETSC_COMM_WORLD,&Sp);CHKERRQ(ierr);
  ierr = MatSetFromOptions(Sp);CHKERRQ(ierr);
  ierr = MatLoad(Sp,viewer);CHKERRQ(ierr);
  ierr = PetscViewerDestroy(&viewer);CHKERRQ(ierr);

  ierr = PetscOptionsGetString(NULL,NULL,"-fMp",filename,PETSC_MAX_PATH_LEN,&flg);CHKERRQ(ierr);
  if (!flg) SETERRQ(PETSC_COMM_WORLD,1,"Must indicate a file name for matrix Mp with the -fMp option");
  ierr = PetscViewerBinaryOpen(PETSC_COMM_WORLD,filename,FILE_MODE_READ,&viewer);CHKERRQ(ierr);
  ierr = MatCreate(PETSC_COMM_WORLD,&Mp);CHKERRQ(ierr);
  ierr = MatSetFromOptions(Mp);CHKERRQ(ierr);
  ierr = MatLoad(Mp,viewer);CHKERRQ(ierr);
  ierr = PetscViewerDestroy(&viewer);CHKERRQ(ierr);

  ierr = PetscOptionsGetString(NULL,NULL,"-fLp",filename,PETSC_MAX_PATH_LEN,&flg);CHKERRQ(ierr);
  if (!flg) SETERRQ(PETSC_COMM_WORLD,1,"Must indicate a file name for matrix Lp with the -fLp option");
  ierr = PetscViewerBinaryOpen(PETSC_COMM_WORLD,filename,FILE_MODE_READ,&viewer);CHKERRQ(ierr);
  ierr = MatCreate(PETSC_COMM_WORLD,&Lp);CHKERRQ(ierr);
  ierr = MatSetFromOptions(Lp);CHKERRQ(ierr);
  ierr = MatLoad(Lp,viewer);CHKERRQ(ierr);
  ierr = PetscViewerDestroy(&viewer);CHKERRQ(ierr);

  KSPCreate(PETSC_COMM_WORLD, &ksp);
  KSPSetOperators(ksp, Sp, Sp);
  KSPSetFromOptions(ksp);
  KSPGetPC(ksp, &pc);
  ierr = PCGetType(pc, &pctype);CHKERRQ(ierr);

  ierr = KSPGetType(ksp, &ksptype);CHKERRQ(ierr);

  if(!strcmp(pctype, PCCOMPOSITE)) {
    PCCompositeAddPC(pc,PCNONE);
    PCCompositeAddPC(pc,PCNONE);
    PC subpc;
    PCCompositeGetPC(pc,1,&subpc);
    PCSetOperators(subpc, Mp, Mp);
    PCSetFromOptions(subpc);
    PCCompositeGetPC(pc,0,&subpc);
    PCSetOperators(subpc, Lp, Lp);
    PCSetFromOptions(subpc);
  }
  else if((!strcmp(pctype, PCSHELL)) && !strcmp(ksptype, KSPPREONLY)) { // PC shell defined with a sequence of user-defined operations
    ierr   = PetscPrintf(PETSC_COMM_SELF,"  Setting pressure PC to PCShell ...\n");CHKERRQ(ierr);
    /* Create a context for the user-defined preconditioner */
    ierr = SampleShellPCCreate(&shell,Mp,Lp);CHKERRQ(ierr);

    /* Set the user-defined routine for applying the preconditioner */
    ierr = PCShellSetApply(pc,SampleShellPCApply);CHKERRQ(ierr);
    ierr = PCShellSetContext(pc,shell);CHKERRQ(ierr);

    /* Set user-defined function to free objects used by custom preconditioner */
    ierr = PCShellSetDestroy(pc,SampleShellPCDestroy);CHKERRQ(ierr);

    /* Set a name for the preconditioner, used for PCView() */
    ierr = PCShellSetName(pc,"MyPreconditioner");CHKERRQ(ierr);

  }
  else if((!strcmp(pctype, PCSHELL)) && strcmp(ksptype, KSPPREONLY)) {
     SETERRQ(PETSC_COMM_WORLD,1,"Here, PCSHELL type must be called with KSPPREONLY ");
  }
  Vec x, b;
  PetscScalar one = 1;
  KSPSetFromOptions(ksp);
  MatCreateVecs(Sp,&x,&b);
  VecSet(b, one);
  VecSet(x, one);
  KSPSolve(ksp, b, x);
  VecView(x, PETSC_VIEWER_STDOUT_SELF);
  VecDestroy(&x);
  VecDestroy(&b);
  KSPDestroy(&ksp);
  ierr = PetscFinalize();
  return ierr;
}

PetscErrorCode SampleShellPCCreate(SampleShellPC **shell, Mat Mp, Mat Lp) {
  SampleShellPC  *newctx;
  PetscErrorCode ierr;
  PC pctmp;

  ierr = PetscNew(&newctx);CHKERRQ(ierr);

  ierr = KSPCreate(PETSC_COMM_WORLD,&newctx->kspSchurMp);CHKERRQ(ierr);
  ierr = KSPSetOperators(newctx->kspSchurMp,Mp,Mp);CHKERRQ(ierr);
  ierr = KSPSetType(newctx->kspSchurMp, KSPPREONLY);CHKERRQ(ierr);
  ierr = KSPGetPC(newctx->kspSchurMp,&pctmp);CHKERRQ(ierr);
  ierr = PCSetType(pctmp, PCJACOBI);CHKERRQ(ierr);

  ierr = KSPCreate(PETSC_COMM_WORLD,&newctx->kspSchurLp);CHKERRQ(ierr);
  ierr = KSPSetOperators(newctx->kspSchurLp,Lp,Lp);CHKERRQ(ierr);
  ierr = KSPSetType(newctx->kspSchurLp, KSPPREONLY);CHKERRQ(ierr);
  ierr = KSPGetPC(newctx->kspSchurLp,&pctmp);CHKERRQ(ierr);
  ierr = PCSetType(pctmp, PCJACOBI);CHKERRQ(ierr);

  *shell = newctx;
  return 0;
}
PetscErrorCode SampleShellPCApply(PC pc,Vec x,Vec y) {
  SampleShellPC  *shell;
  PetscErrorCode ierr;
  Vec *tmp;

  ierr = VecZeroEntries(y);CHKERRQ(ierr);
  ierr = VecDuplicateVecs(y,2,&tmp);

  ierr = PCShellGetContext(pc,(void**)&shell);CHKERRQ(ierr);
  ierr = KSPSolve(shell->kspSchurMp,x,tmp[0]);CHKERRQ(ierr);
  ierr = KSPSolve(shell->kspSchurLp,x,tmp[1]);CHKERRQ(ierr);
  PetscScalar one = 1.0;
  ierr = VecAXPY(y,one,tmp[0]);CHKERRQ(ierr);
  ierr = VecAXPY(y,one,tmp[1]);CHKERRQ(ierr);
  return 0;
}
PetscErrorCode SampleShellPCDestroy(PC pc) {
  SampleShellPC  *shell;
  PetscErrorCode ierr;

  ierr = PCShellGetContext(pc,(void**)&shell);CHKERRQ(ierr);
  ierr = KSPDestroy(&shell->kspSchurMp);CHKERRQ(ierr);
  ierr = KSPDestroy(&shell->kspSchurLp);CHKERRQ(ierr);
  ierr = PetscFree(shell);CHKERRQ(ierr);
  return 0;
}
